import json
import random
from typing import Any, Self, get_args

from attr import define, fields, field

from .DbClassCreator import DbClassCreator
from .db_attrs_converter import db_attrs_converter
from .JsonEncoder import Decoder
from .JsonEncoder.default_json_encoder import json_encoder


@define
class DbClass(metaclass=DbClassCreator):
    _id = field(init=False, type=Any, factory=lambda: random.randint(-2**63, 2**63 - 1))

    def __attrs_post_init__(self):
        if hasattr(self, 'id'):
            self._id = self.id
        self._decode()

    def serialize(self) -> dict:
        from .DbClassLiteral import DbClassLiteral

        return json.loads(
            json.dumps(
                dict(
                    (
                        f.name,
                        getattr(self, f.name)._id
                        if isinstance(getattr(self, f.name), DbClass)
                        and not isinstance(getattr(self, f.name), DbClassLiteral)
                        else getattr(self, f.name),
                    )
                    for f in fields(type(self))
                ),
                cls=json_encoder,
            )
        )

    @classmethod
    def deserialize(cls, dictionary: dict) -> Self:
        deserialized = db_attrs_converter.structure(dictionary, cls)
        deserialized._fill_id(dictionary)
        return deserialized

    def _fill_id(self, dictionary: dict):
        from .DbClassLiteral import DbClassLiteral

        self._id = dictionary["_id"]
        for f in fields(type(self)):
            types = list(get_args(f.type))
            if isinstance(f.type, type):
                types.append(f.type)
            if any(issubclass(type(self) if field_type == Self else field_type, DbClassLiteral) for field_type in types):
                f.type._fill_id(getattr(self, f.name), dictionary[f.name])

    def _decode(self):
        for f in fields(type(self)):
            for decoder in Decoder.__subclasses__():
                if decoder.is_valid(f.type):
                    setattr(self, f.name, decoder.decode(getattr(self, f.name), f.type))
                    break



__all__ = [
    "DbClass",
    "DbClassLiteral",
    "db_attrs_converter",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "varchar",
    "text",
]

from .DbClassLiteral import DbClassLiteral
from .DbClass import DbClass
from .db_attrs_converter import db_attrs_converter
from .db_fields.texts import char, varchar, text
from .db_fields.ints import int8, int16, int32, int64, uint8, uint16, uint32, uint64



from functools import partial

from attr import field, validators, NOTHING


def _check_char(instance, attribute, value, length: int):
    if len(value) != length:
        raise ValueError(f"Length of {value=} must be {length} is {len(value)}")


def _check_varchar(instance, attribute, value, max_length: int):
    if len(value) > max_length:
        raise ValueError(
            f"Length of {value=} must be less than {max_length} is {len(value)}"
        )


def char(
    length: int,
    *,
    default=NOTHING,
    validator=None,
    repr=True,
    hash=None,
    init=True,
    metadata=None,
    type=None,
    converter=None,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=None,
    alias=None,
):
    return partial(
        field,
        validator=[validators.instance_of(str), partial(_check_char, length=length)]
        if validator is None
        else validator,
    )(
        default=default,
        repr=repr,
        hash=hash,
        init=init,
        metadata=metadata,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
        alias=alias,
        type=type,
    )


def varchar(
    max_length: int,
    *,
    default=NOTHING,
    validator=None,
    repr=True,
    hash=None,
    init=True,
    metadata=None,
    type=None,
    converter=None,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=None,
    alias=None,
):
    return partial(
        field,
        validator=[
            validators.instance_of(str),
            partial(_check_varchar, max_length=max_length),
        ]
        if validator is None
        else validator,
    )(
        default=default,
        repr=repr,
        hash=hash,
        init=init,
        metadata=metadata,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
        alias=alias,
        type=type,
    )


text = partial(field, validator=[validators.instance_of(str)])



import json
from typing import Any

from attr import has, asdict

from ...JsonEncoder import Encoder


class AttrsEncoder(Encoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return has(element)

    @staticmethod
    def encode(element) -> dict:
        from ..DefaultJsonEncoder import DefaultJsonEncoder
        from ... import DbClassLiteral, DbClass

        if isinstance(element, DbClassLiteral) or not isinstance(element, DbClass):
            return json.loads(json.dumps(asdict(element), cls=DefaultJsonEncoder))
        return element._id



from abc import abstractmethod
from typing import Any

from ..JsonEncoder.JsonOperator import JSONOperator


class Decoder(JSONOperator):
    @staticmethod
    @abstractmethod
    def is_valid(element: Any) -> bool:
        pass

    @staticmethod
    @abstractmethod
    def decode(element: Any, _) -> str:
        pass



from datetime import datetime

from ...JsonEncoder import Decoder


class DatetimeDecoder(Decoder):
    @staticmethod
    def is_valid(field_type: type) -> bool:
        return field_type == datetime

    @staticmethod
    def decode(element: str, _) -> datetime:
        if isinstance(element, datetime):
            return element
        return datetime.fromtimestamp(float(element))





from json import JSONEncoder

from .Encoder import Encoder


class DefaultJsonEncoder(JSONEncoder):
    def default(self, obj):
        for encoder in Encoder.__subclasses__():
            if encoder.is_valid(obj):
                return encoder.encode(obj)
        else:
            raise ValueError(obj + " can't be encoded")



from decimal import Decimal
from typing import Any

from ...JsonEncoder import Decoder


class DecimalDecoder(Decoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return element == Decimal

    @staticmethod
    def decode(element: Any, _) -> Decimal:
        return Decimal(element)



from decimal import Decimal
from typing import Any

from ...JsonEncoder.Encoder import Encoder


class DecimalEncoder(Encoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return isinstance(element, Decimal)

    @staticmethod
    def encode(element: Decimal) -> str:
        return str(element)



from typing import get_args, Self

from attr import fields
from attrs import define

from .DbClass import DbClass
from .db_attrs_converter import db_attrs_converter


@define
class DbClassLiteral(DbClass):
    pass


def _handle_new_db(value, db_class_type):
    if isinstance(value, int):
        return value
    id_value = value.pop("_id")
    new_instance = db_class_type(**value)
    new_instance._id = id_value
    value["_id"] = id_value
    for f in fields(db_class_type):
        types = list(get_args(f.type))
        if isinstance(f.type, type):
            types.append(f.type)
        references = tuple(field_type for field_type in types if isinstance(field_type, str))
        if references:
            raise ValueError("Data can't be deserialized as field {} in {} has a forward references {}. Consider making classes {} DbClasses and initializing them.".format(f.name, db_class_type.__name__, references, references))
        try:
            if any(issubclass(db_class_type if field_type == Self else field_type, DbClassLiteral) for field_type in types):
                setattr(
                    new_instance,
                    f.name,
                    db_attrs_converter.structure(value[f.name], f.type),
                )
        except:
            pass
    return new_instance


db_attrs_converter.register_structure_hook_func(
    lambda t: issubclass(t, DbClass), _handle_new_db
)



from functools import partial

from attr import field, validators


def _check_int(instance, attribute, value, bits: int):
    if value < -(2 ** (bits - 1)):
        raise ValueError(f"{value=} must be more than {-(2 ** (bits - 1))}")
    if value > 2 ** (bits - 1) - 1:
        raise ValueError(f"{value=} must be less than {2 ** (bits - 1) - 1}")


def _check_uint(instance, attribute, value, bits: int):
    if value < 0:
        raise ValueError(f"{value=} must be positive")
    if value > 2**bits - 1:
        raise ValueError(f"{value=} must be less than {2 ** bits - 1}")


int8 = partial(
    field, validator=[validators.instance_of(int), partial(_check_int, bits=8)]
)
int16 = partial(
    field, validator=[validators.instance_of(int), partial(_check_int, bits=16)]
)
int32 = partial(
    field, validator=[validators.instance_of(int), partial(_check_int, bits=32)]
)
int64 = partial(
    field, validator=[validators.instance_of(int), partial(_check_int, bits=64)]
)
uint8 = partial(
    field, validator=[validators.instance_of(int), partial(_check_uint, bits=8)]
)
uint16 = partial(
    field, validator=[validators.instance_of(int), partial(_check_uint, bits=16)]
)
uint32 = partial(
    field, validator=[validators.instance_of(int), partial(_check_uint, bits=32)]
)
uint64 = partial(
    field, validator=[validators.instance_of(int), partial(_check_uint, bits=64)]
)



from enum import Enum
from typing import Any

from ...JsonEncoder.Encoder import Encoder


class EnumEncoder(Encoder):
    @staticmethod
    def is_valid(element: Any) -> bool:
        return isinstance(element, Enum)

    @staticmethod
    def encode(element: Enum) -> str:
        from ..DefaultJsonEncoder import DefaultJsonEncoder

        return eval(DefaultJsonEncoder().encode(element.value))
